/************************************************************************/
/*  Routines to get information about the input stream.			*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2013 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static const char sccsid[] = "$Id: get_stream_info.c,v 1.9 2013/04/30 19:42:06 doug Exp $ ";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <math.h>

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "qlib2.h"

#include "params.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"

/************************************************************************/
/*  get_stream_info:							*/
/*	Get information about the input data stream.			*/
/************************************************************************/
int get_stream_info 
   (ST_INFO	*st_p)		/* ptr to current stream header.	*/
{
    STORE_FILE_HEAD *sfh;
    STORE_DATA	    *sd;
    QDA_HDR	    *qh;
    SDR_HDR	    *sh;
    DATA_HDR	    *hdr;
    int		    seq_no;
    int		    skip;
    char	    *hdrtype_string = NULL;

    st_p->blksize = 0;
    if (fill_buffer (st_p, MAX_BLKSIZE) == 0) return (EOF);

    /*	Determine the type of the file, and fill in header information.	*/
    /*	Get:								*/
    /*	    header_type, station, stream, sample rate, and blksize.	*/

    st_p->hdr_type = get_header_type (st_p->iob->buf, st_p->iob->free);

    switch (st_p->hdr_type) {
    case DRM_HDR_TYPE:
	hdrtype_string = "DRM_HDR";
	sfh = (STORE_FILE_HEAD *)(st_p->iob->buf);
	strcpy(st_p->station, sfh->stream_config.station_name);
	strcpy(st_p->channel, sfh->stream_config.seedchannel);
	strcpy(st_p->network, "  ");
	strcpy(st_p->location, "  ");
	trim (st_p->station);
	trim (st_p->channel);
	trim (st_p->network);
	trim (st_p->location);
	st_p->blksize =  sfh->store_pack_size;
	if (st_p->blksize != sizeof (STORE_DATA)) {
	    fprintf (info, "Warning: blksize says %d, should be %d\n", st_p->blksize,
		     (int)sizeof (STORE_DATA));
	    fflush (info);
	}
	if (sfh->store_file_type != STYPE_CONT) {
	    fprintf (info, "Error: filetype = %d, should be %d\n", 
		     sfh->store_file_type, STYPE_CONT);
	    fflush (info);
	    exit(1);
	}
	st_p->begtime = decode_time_drm (sfh->store_begin,SEED_BIG_ENDIAN);
	st_p->endtime = decode_time_drm (sfh->store_end,SEED_BIG_ENDIAN);

	/* Read the first block to get the rest of the info.		*/
	/* Rate is ONLY available in the block header.			*/
	/* The first block begins at offset				*/
	/*	store_offset*store_pack_size				*/
	/* in the file.							*/
	st_p->iob->avail = sizeof(STORE_FILE_HEAD);
	skip = (sfh->store_offset * sfh->store_pack_size)- sizeof(STORE_FILE_HEAD);
	if (fill_buffer (st_p, skip) < skip) return(EOF);
	st_p->iob->avail = skip;
	if (fill_buffer (st_p, st_p->blksize) < st_p->blksize) return(EOF);
	sd = (STORE_DATA *)(st_p->iob->buf);
	st_p->rate = sd->rate;
	st_p->rate_mult = 1;
	break;

    case QDA_HDR_TYPE:
	hdrtype_string = "QDA_HDR";
	qh = (QDA_HDR *)(st_p->iob->buf);
	hdr = decode_hdr_qda (qh, st_p->iob->free);
	if (hdr == NULL) {
	    fprintf (info, "Error: Unable to decode volume hdr for %s, qlib2_errno = %d\n",
		     st_p->filename, qlib2_errno);
	    fflush (info);
	    exit(1);
	}
	strcpy (st_p->station, hdr->station_id);
	strcpy (st_p->channel, hdr->channel_id);
	strcpy (st_p->network, hdr->network_id);
	strcpy (st_p->location, hdr->location_id);
	st_p->rate = hdr->sample_rate;
	st_p->rate_mult = 1;
	seq_no = hdr->seq_no;
	st_p->blksize = hdr->blksize;
	if (st_p->blksize == 0) {
	    fprintf (stderr, "Error: invalid blocksize for %s\n", st_p->filename);
	    exit(1);
	}
	free_data_hdr (hdr);
	break;

    case SDR_VOL_HDR_TYPE:
	hdrtype_string = "VOH_HDR";
	sh = (SDR_HDR *)(st_p->iob->buf);
	hdr = decode_hdr_sdr(sh,st_p->iob->free);
	if (hdr == NULL) {
	    fprintf (info, "Error: Unable to decode volume hdr for %s, qlib2_errno = %d\n",
		     st_p->filename, qlib2_errno);
	    fflush (info);
	    exit(1);
	}
	st_p->blksize = hdr->blksize;
	if ( (st_p->vfh = (char *)malloc(st_p->blksize+1)) == NULL) {
	    fprintf (info, "Error: unable to malloc vfh structure.\n");
	    fflush (info);
	    exit(1);
	}
	memcpy ((void *)st_p->vfh, (void *)sh, st_p->blksize);
	*((st_p->vfh) + st_p->blksize) = '\0';

	/* The volume blockette number is in the data_type field.   */
	if (hdr->data_type > 0) {
	    BS *bs;	/* ptr to blockette structure.		*/
	    char *vb;	/* ptr to volume blockette.		*/
	    INT_TIME *pt;
	    char date_str[40];
	    char *p;
	    int i, j;
	    unsigned char type;

	    if (!(bs=find_blockette (hdr, hdr->data_type))) {
		fprintf (info, "Error: cannot find volume blockette\n");
		fflush (info);
		exit(1);
	    }
	    vb = (char *)(bs->pb);

	    /* Blocksize was already parsed by decode_hdr function */
	    type = hdr->data_type;
	    switch (type) {
	      case 5:		/* field volume */
	      case 10:		/* volume id */
		break;
	      case 8:		/* telemetry volume */
		trim(charncpy(st_p->station,vb+13,5));
		trim(charncpy(st_p->location,vb+18,2));
		trim(charncpy(st_p->channel,vb+20,3));
		/* Skip 4 variable strings */
		p = vb+23;
		i = 4;
		j = bs->len;
		while (i > 0 && j-- > 0) {
		    if (*p++ == '~') --i;
		}
		/* Copy network if not at the end of the volume blockette. */
		if (j > 0) trim(charncpy(st_p->network,p,2));

		/* Get volume begtime and endtime. */
		charvncpy(date_str,vb+23,22,0);
		if ((pt = parse_date(date_str)) != NULL) st_p->begtime = *pt;
		else fprintf (info, "Unable to get start time from volume hdr\n");
		charvncpy(date_str,vb+23,22,1);
		if ((pt = parse_date(date_str)) != NULL) st_p->endtime = *pt;
		else fprintf (info, "Unable to get end time from volume hdr\n");
		fflush (info);
		break;
	      default:
		fprintf (info, "Error: Unknown volume blockette %d\n", hdr->data_type);
		fflush (info);
		exit(1);
	    }
	}

	/* Read the first data block to get the rest of the info.	*/
	st_p->iob->avail = st_p->blksize;
	if (fill_buffer (st_p, st_p->blksize) < st_p->blksize) return(EOF);
	/* Fall through and pick up info from first SEED data record. */
	st_p->hdr_type = SDR_HDR_TYPE;
	free_data_hdr (hdr);

    case SDR_HDR_TYPE:
	if (hdrtype_string == NULL) hdrtype_string = "SDR_HDR";
	sh = (SDR_HDR *)(st_p->iob->buf);
	hdr = decode_hdr_sdr(sh,st_p->iob->free);
	if (hdr == NULL) {
	    fprintf (info, "Error: Unable to decode data hdr for %s, qlib2_errno = %d\n",
		     st_p->filename, qlib2_errno);
	    fflush (info);
	    exit(1);
	}
	strcpy (st_p->station, hdr->station_id);
	strcpy (st_p->channel, hdr->channel_id);
	strcpy (st_p->network, hdr->network_id);
	strcpy (st_p->location, hdr->location_id);
	trim (st_p->station);
	trim (st_p->channel);
	trim (st_p->network);
	trim (st_p->location);
	st_p->rate = hdr->sample_rate;
	st_p->rate_mult = hdr->sample_rate_mult;
	st_p->blksize = hdr->blksize;
	st_p->begtime = hdr->begtime;
	free_data_hdr(hdr);
	if (st_p->blksize > MAX_BLKSIZE) {
	    fprintf(stderr, "Error: Block size %d is greater than MAX_BLKSIZE %d\n",
		    st_p->blksize, MAX_BLKSIZE);
	    exit(1);
	}
	if (st_p->blksize == 0) {
	    fprintf (stderr, "Error: invalid blocksize for %s\n", st_p->filename);
	    exit(1);
	}

	/* If the input stream is seekable, get endtime from the last record.	*/
	struct stat stat;
	if (missing_time(st_p->endtime) && (st_p->iob->pos = ftell (st_p->iob->fp)) >= 0 &&
	    fstat(fileno(st_p->iob->fp), &stat) == 0) {
	    IOB *save_iob = st_p->iob;
	    st_p->iob = new_iob();
	    st_p->iob-> fp = save_iob->fp;
	    off_t newpos = ((st_p->streamlimit >= 0) ? st_p->streamlimit : stat.st_size) - st_p->blksize;
	    /* Reposition file to last block and get endtime. */
	    if (fseeko (st_p->iob->fp, newpos, SEEK_SET) == 0 && fill_buffer (st_p, st_p->blksize) == st_p->blksize) {
		DATA_HDR *save_hdr = st_p->cur_hdr;
		st_p->cur_hdr = NULL;
		get_block_info(st_p);
		if (st_p->cur_hdr) {
		    st_p->endtime = st_p->cur_hdr->endtime;
		    free_data_hdr (st_p->cur_hdr);
		    --(st_p->num_blocks);
		}
		st_p->cur_hdr = save_hdr;
	    }
	    st_p->iob->fp = NULL;
	    free_iob(st_p->iob);
	    st_p->iob = save_iob;
	    if (fseeko (st_p->iob->fp, st_p->iob->pos, SEEK_SET) != 0) {
		fprintf (info, "Error: unable to re-position file %s after getting endtime\n", st_p->filename);
		fflush (info);
		exit(1);
	    }
	}
	break;

    default:
	fprintf (stderr, "Error: unrecognizable input file %s\n", st_p->filename);
	exit(1);
	break;
    }

    if (debug(DEBUG_FILELIMIT) && (! missing_time(st_p->begtime)) && (! missing_time(st_p->endtime))) {
	char *begstr = strdup(time_to_str(st_p->begtime, date_fmt));
	char *endstr = strdup(time_to_str(st_p->endtime, date_fmt));
	fprintf(info, "file %s unit %d limits begtime=%s endtime=%s from %s\n", 
		st_p->filename, fileno(st_p->iob->fp), begstr, endstr, hdrtype_string);
	fflush (info);
	free(begstr);
	free(endstr);
    }
			
    return(0);
}

/************************************************************************/
/*  get_header_type:							*/
/*	Determine the type of input stream.				*/
/************************************************************************/
int get_header_type 
   (char    *p,			/* ptr to buffer containing header.	*/
    int	    nr)			/* max number of bytes for header.	*/
{
    if (is_drm_header((STORE_FILE_HEAD *)p,nr)) return DRM_HDR_TYPE;
    else if (is_qda_header((QDA_HDR *)p,nr)) return QDA_HDR_TYPE;
    else if (is_sdr_header((SDR_HDR *)p,nr)) return SDR_HDR_TYPE;
    else if (is_sdr_vol_header((SDR_HDR *)p,nr)) return (SDR_VOL_HDR_TYPE);
    else return (-1);
}
